Option Strict Off
Imports System.ComponentModel
Imports Bradford.Common.ExceptionHandling
Imports Bradford.Win.ControlLibrary.DatabaseObject
Imports Infragistics.Win.UltraWinGrid.ExcelExport
Imports Infragistics.Win.UltraWinGrid
Imports Bradford.Common
Imports Bradford.ObjectLibrary
Imports Infragistics.Win
Imports System.Text.RegularExpressions
Imports Infragistics.Documents.Excel


Public Class BSCUltraDataGrid
    Public Enum BSCUltraDataGridStyle
        ''' <summary>
        ''' Use Existing Layout
        ''' </summary>
        ''' <remarks></remarks>
        UltraGrid = 0
        ''' <summary>
        ''' Read Only Browse Multiple Selections Allowed.
        ''' </summary>
        ''' <remarks></remarks>
        BrowseMultiSelect = 3
        ''' <summary>
        ''' Read Only Browse Single Selection Allowed.
        ''' </summary>
        ''' <remarks></remarks>
        BrowseSingleSelect = 2
        ''' <summary>
        ''' Custom Layout (Plug-in): Not Implemented
        ''' </summary>
        ''' <remarks></remarks>
        Custom = 1

        ''' <summary>
        ''' Standard Look and Feel
        ''' </summary>
        ''' <remarks></remarks>
        Standard = 4
    End Enum

    Private _ExManager As BSCExceptionHandler
    Private _HiddenColumns As Dictionary(Of String, HiddenColumnItem)
    Private _ExcelMaxRows As Integer

    Public Sub New()

        ' This call is required by the Windows Form Designer.
        InitializeComponent()

        Try
            _SelectedRows = New List(Of UltraGridRow)

            If Not Me.DesignMode Then
                ' Add any initialization after the InitializeComponent() call.
                _ExManager = New BSCExceptionHandler
                _ExcelExporter = New Infragistics.Win.UltraWinGrid.ExcelExport.UltraGridExcelExporter
                _ExcelExporter.FileLimitBehaviour = FileLimitBehaviour.TruncateData

                BuildContextMenu()
                BuildSmartTagMenuLeft()

                _HiddenColumns = New Dictionary(Of String, HiddenColumnItem)(StringComparer.CurrentCultureIgnoreCase)
            End If

        Catch ex As Exception
            MessageBox.Show(ex.ToString, Me.Name & ".New() Exception", MessageBoxButtons.OK, MessageBoxIcon.Stop)
        End Try
    End Sub


#Region "Events"
    ''' <summary>
    ''' Start Export Event (after filename is set).
    ''' </summary>
    ''' <param name="fileName">Export file name.</param>
    ''' <param name="sender">Object</param>
    ''' <param name="e">EventArg to cancel export </param>
    ''' <remarks></remarks>
    Public Event BeforeExport(ByVal fileName As String, ByRef sender As Object, ByRef e As Bradford.Common.CancelEventArg)
    Public Event ExportAborted As EventHandler
    ''' <summary>
    ''' Export Completed.
    ''' </summary>
    ''' <param name="fileName">Exported file name.</param>
    ''' <param name="sender">Object</param>
    ''' <param name="e">EventArt to cancel displaying exported file name.</param>
    ''' <remarks></remarks>
    Public Event AfterExport(ByVal fileName As String, ByRef sender As Object, ByVal e As Bradford.Common.CancelEventArg)
    Public Event Progress(ByVal current As Integer, ByVal total As Integer)
    Public Event Rebind()

    Public Event BeforeLoadGrid(ByVal sender As Object, ByVal e As CancelEventArg)
    Public Event AfterLoadGrid As EventHandler
    Public Event ExecuteLoadGrid()

    ''' <summary>
    ''' Event before row is added to the custom selected rows list.
    ''' </summary>
    ''' <param name="row">Row</param>
    ''' <remarks></remarks>
    Public Event BeforeSelectRow(ByVal row As UltraGridRow)
    ''' <summary>
    ''' Event after row is added to the custom selected rows list.
    ''' </summary>
    ''' <param name="row">Row</param>
    ''' <remarks></remarks>
    Public Event AfterSelectRow(ByVal row As UltraGridRow)
    ''' <summary>
    ''' Event before row is removed from the custom selected rows list.
    ''' </summary>
    ''' <param name="row">Row</param>
    ''' <remarks></remarks>
    Public Event BeforeRemoveRow(ByVal row As UltraGridRow)
    ''' <summary>
    ''' Event after row is removed from the custom selected rows list.
    ''' </summary>
    ''' <param name="row">Row</param>
    ''' <remarks></remarks>
    Public Event AfterRemoveRow(ByVal row As UltraGridRow)

    ''' <summary>
    ''' Occurs before initializing current layout.
    ''' </summary>
    ''' <developer>Pete Sral</developer>
    Public Event BeforeInitializeCurrentLayout(ByVal sender As Object, ByVal e As CancelEventArg, ByVal layout As UltraGridLayout)

    ''' <summary>
    ''' Occurs when after current layout initialized.
    ''' </summary>
    ''' <developer>Pete Sral</developer>
    Public Event AfterInitializeCurrentLayout(ByVal sender As Object, ByVal e As EventArgs, ByVal layout As UltraGridLayout)

#End Region

#Region "OnEvent Methods"
    ''' <summary>
    ''' Event before row is added to the custom selected rows list.
    ''' </summary>
    ''' <param name="row">Row</param>
    ''' <remarks></remarks>
    Protected Friend Overridable Sub OnBeforeSelectRow(ByVal row As UltraGridRow)
        RaiseEvent BeforeSelectRow(row)
    End Sub
    ''' <summary>
    ''' Event after row is added to the custom selected rows list.
    ''' </summary>
    ''' <param name="row">Row</param>
    ''' <remarks></remarks>
    Protected Friend Overridable Sub OnAfterSelectedRow(ByVal row As UltraGridRow)
        RaiseEvent AfterSelectRow(row)
    End Sub
    ''' <summary>
    ''' Event before row is removed from the custom selected rows list.
    ''' </summary>
    ''' <param name="row">Row</param>
    ''' <remarks></remarks>
    Protected Friend Overridable Sub OnBeforeRemoveRow(ByVal row As UltraGridRow)
        RaiseEvent BeforeRemoveRow(row)
    End Sub
    ''' <summary>
    ''' Event after row is removed from the custom selected rows list.
    ''' </summary>
    ''' <param name="row">Row</param>
    ''' <remarks></remarks>
    Protected Friend Overridable Sub OnAfterRemoveRow(ByVal row As UltraGridRow)
        RaiseEvent AfterRemoveRow(row)
    End Sub


#End Region


#Region "Property"
    Private _ResetLayoutAfterLoad As Boolean
    ''' <summary>
    ''' Gets or sets a value indicating whether to reset layout after each data load.
    ''' </summary>
    ''' <value>
    ''' <c>true</c> if [reset layout after load]; otherwise, <c>false</c>.
    ''' </value>
    ''' <developer>Pete Sral</developer>
    <DefaultValue(False)> _
    <Description("Reset layout after each data load.")> _
    Public Property ResetLayoutAfterLoad() As Boolean
        Get
            Return _ResetLayoutAfterLoad
        End Get
        Set(ByVal value As Boolean)
            _ResetLayoutAfterLoad = value
        End Set
    End Property

    Private _AutoResizeAllRowCount As Integer = 1000
    ''' <summary>
    ''' Gets or sets the auto resize all row count.
    ''' </summary>
    ''' <value>The auto resize all row count.</value>
    ''' <developer>Pete Sral</developer>           
    <DefaultValue(1000)> _
    Public Property AutoResizeAllRowCount() As Integer
        Get
            Return _AutoResizeAllRowCount
        End Get
        Set(ByVal value As Integer)
            _AutoResizeAllRowCount = value
        End Set
    End Property

    ''' <summary>
    ''' Global and DB object is initialized
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public ReadOnly Property IsInitialized() As Boolean
        Get
            Return GlobalObject IsNot Nothing AndAlso DB IsNot Nothing
        End Get
    End Property
    Private _DB As DatabaseObject.BSCUltraDataGridDB
    ''' <summary>
    ''' Database Object
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public ReadOnly Property DB() As BSCUltraDataGridDB
        Get
            Return _DB
        End Get
    End Property

    Private _GlobalObject As Bradford.ObjectLibrary.IGlobalObjectBase
    <Browsable(False)> _
    Public ReadOnly Property GlobalObject() As IGlobalObjectBase
        Get
            Return _GlobalObject
        End Get
    End Property



    Private _ProgressBar As ToolStripProgressBar
    ''' <summary>
    ''' Linked progress bar.
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>

    Public Property ProgressBar() As ToolStripProgressBar
        Get
            Return _ProgressBar
        End Get
        Set(ByVal value As ToolStripProgressBar)
            _ProgressBar = value
        End Set
    End Property

    Private _UniqueContainerName As String
    ''' <summary>
    ''' Gets or sets the name of the unique container.
    ''' </summary>
    ''' <value>The name of the unique container.</value>
    ''' <developer>Pete Sral</developer>
    Public Property UniqueContainerName() As String
        Get
            If String.IsNullOrEmpty(_UniqueContainerName) AndAlso Not DesignMode Then
                _UniqueContainerName = GenerateUniqueContainerName()
            End If
            Return _UniqueContainerName
        End Get
        Set(ByVal value As String)
            _UniqueContainerName = value
        End Set
    End Property



    Private _GridLayoutStyle As BSCUltraDataGridStyle = BSCUltraDataGridStyle.UltraGrid
    ''' <summary>
    ''' Defined
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Property GridLayoutStyle() As BSCUltraDataGridStyle
        Get
            Return _GridLayoutStyle
        End Get
        Set(ByVal value As BSCUltraDataGridStyle)
            _GridLayoutStyle = value
        End Set
    End Property

#End Region

#Region "Event Handlers"
    Private Sub EventHandlerRefresh(ByVal sender As Object, ByVal e As EventArgs)
        RaiseEvent Rebind()
    End Sub

    Private Sub EventHandlerPrintPreview(ByVal sender As Object, ByVal e As EventArgs)
        MessageBox.Show("Print Preview")
    End Sub

    Private Sub EventHandlerPrintPrinter(ByVal sender As Object, ByVal e As EventArgs)
        MessageBox.Show("Print Printer")
    End Sub
    Private Sub EventHandlerExport(ByVal sender As Object, ByVal e As EventArgs)
        ExportExcel()
    End Sub

    Private Sub EventHandlerColumns(ByVal sender As Object, ByVal e As EventArgs)


    End Sub
#End Region

#Region "Grid Events"

    Protected Friend Overridable Sub HandleAfterSelectChange(ByVal sender As Object, ByVal e As Infragistics.Win.UltraWinGrid.AfterSelectChangeEventArgs) Handles Me.AfterSelectChange
        If SelectedRows.Count = 0 Then
            Exit Sub
        End If

        For Each exRow As UltraGridRow In SelectedRows
            If Not Me.Selected.Rows.Contains(exRow) Then
                Me.Selected.Rows.Add(exRow)
            End If
        Next
    End Sub
    Private Sub BSCUltraDataGrid_Enter(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.Enter
        ShowSmartTagButton()
    End Sub

    Private Sub BSCUltraDataGrid_InitializeLayout(ByVal sender As Object, ByVal e As Infragistics.Win.UltraWinGrid.InitializeLayoutEventArgs) Handles Me.InitializeLayout

        DefaultLayout = e.Layout

        If Not HasSavedLayout Then
            InitializeCurrentLayout(sender, e.Layout)
            CurrentLayout = e.Layout
        End If
    End Sub

    ''' <summary>
    ''' Initializes the current layout.
    ''' </summary>
    ''' <param name="sender">The sender.</param>
    ''' <param name="layout">The layout.</param>
    ''' <developer>Pete Sral</developer>
    Public Overridable Sub InitializeCurrentLayout(ByVal sender As Object, ByVal layout As UltraGridLayout)
        Try
            If layout Is Nothing Then
                Exit Sub
            End If

            Dim e As New CancelEventArg

            RaiseEvent BeforeInitializeCurrentLayout(sender, e, layout)

            If e.Cancel Then
                Exit Sub
            End If

            'hide saved columns, if hide columns is called before data source is set.
            For Each oItem As KeyValuePair(Of String, HiddenColumnItem) In _HiddenColumns
                If String.IsNullOrEmpty(oItem.Value.ColumnName) Then
                    HideColumn(oItem.Value.Index, True, oItem.Value.Layout, oItem.Value.Band)
                Else
                    HideColumn(oItem.Value.ColumnName, True, oItem.Value.Layout, oItem.Value.Band)
                End If
            Next

            Dim altColor As Color = Color.LightBlue
            'set alt. row coloring
            If GlobalObject IsNot Nothing Then
                altColor = GlobalObject.Settings.GetSetting("GridColorOddNET", Color.LightSkyBlue)
            End If

            For i As Integer = 0 To layout.Bands.Count - 1

                layout.Bands(i).Override.CellMultiLine = DefaultableBoolean.True

                With layout.Bands(i).Override.RowAlternateAppearance
                    .ThemedElementAlpha = Infragistics.Win.Alpha.UseAlphaLevel
                    .AlphaLevel = 255
                    .BackColorAlpha = Infragistics.Win.Alpha.UseAlphaLevel
                    .AlphaLevel = 255
                    .ForegroundAlpha = Infragistics.Win.Alpha.UseAlphaLevel
                    .AlphaLevel = 255
                    .BackColor = altColor
                End With

                'Update column properties
                For Each oCol As UltraGridColumn In layout.Bands(i).Columns
                    With oCol
                        'set caption
                        .Header.Caption = .Header.Caption.Replace("_", "")
                        .Header.Caption = Regex.Replace(.Header.Caption, "([a-z](?=[A-Z0-9])|[A-Z](?=[A-Z][a-z]))", "$1 ")

                        .CellMultiLine = DefaultableBoolean.True
                        .CellAppearance.TextTrimming = TextTrimming.EllipsisCharacter
                    End With
                Next

            Next i


            Select Case GridLayoutStyle
                Case BSCUltraDataGridStyle.BrowseMultiSelect, BSCUltraDataGridStyle.BrowseSingleSelect
                    InitializeLayoutBrowse(sender, layout)
                Case BSCUltraDataGridStyle.Standard
                    InitializeLayoutStandard(sender, layout)
                Case BSCUltraDataGridStyle.Custom
                    'TODO add ability to plug in a class to do init. layout.
            End Select

            'auto size row height to fit data.
            layout.Override.RowSizing = Infragistics.Win.UltraWinGrid.RowSizing.AutoFree

            RaiseEvent AfterInitializeCurrentLayout(sender, e, layout)
        Catch ex As Exception
            Throw New Exception(Me.Name & " - InitializeCurrentLayout.", ex)
        End Try
    End Sub
    Private Sub BSCUltraDataGrid_InitializeRow(ByVal sender As Object, ByVal e As Infragistics.Win.UltraWinGrid.InitializeRowEventArgs) Handles Me.InitializeRow
        ColorSelectedRow(e.Row)
    End Sub

    Private Sub BSCUltraDataGrid_Leave(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.Leave
        HideSmartTagButton()
    End Sub
    Private Sub BSCDataGrid_MouseDown(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseDown
        'show context menu on right mouse click.
        If e.Button = Windows.Forms.MouseButtons.Right And Not GridContextMenu Is Nothing Then
            Me.GridContextMenu.Show(Me, New Point(e.X, e.Y))
        End If
    End Sub

    Private Sub BSCUltraDataGrid_MouseUp(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseUp
        If e.Button = Windows.Forms.MouseButtons.Left Then

            'skip if not in multi row select mode.
            If GridLayoutStyle = BSCUltraDataGridStyle.BrowseSingleSelect OrElse (Me.CurrentLayout IsNot Nothing AndAlso (Me.CurrentLayout.Override.SelectTypeRow = SelectType.Single Or _
               Me.CurrentLayout.Override.SelectTypeRow = SelectType.SingleAutoDrag Or _
               Me.CurrentLayout.Override.SelectTypeRow = SelectType.None Or _
                 Me.CurrentLayout.Override.SelectTypeGroupByRow = Infragistics.Win.UltraWinGrid.SelectType.Single)) Then
                'do not add to list
            Else
                HandleRowSelect(GetRowOnSelectorClick(e))
            End If
        End If
    End Sub

    Private Sub BSCUltraDataGrid_Paint(ByVal sender As Object, ByVal e As System.Windows.Forms.PaintEventArgs) Handles Me.Paint
        If Me.Focused Then
            ShowSmartTagButton()
        End If
    End Sub
#End Region


#Region "Public Methods"
    ''' <summary>
    ''' Handles the exception.
    ''' </summary>
    ''' <param name="ex">The ex.</param>
    ''' <param name="message">The message.</param>
    ''' <developer>Pete Sral</developer>
    Public Overridable Sub HandleEx(ByVal ex As Exception, ByVal message As String)
        If GlobalObject IsNot Nothing Then
            GlobalObject.HandleEx(ex, message)
        Else
            ShowMessage(ex.Message.ToString)
            MessageBox.Show(ex.ToString, ex.Message, MessageBoxButtons.OK, MessageBoxIcon.Error)
        End If
    End Sub
    ''' <summary>
    ''' Reinitialize the grid.
    ''' </summary>
    ''' <param name="globalObj">The global obj.</param>
    ''' <developer>Pete Sral</developer>
    Public Sub ReInitialize(ByRef globalObj As Bradford.ObjectLibrary.IGlobalObjectBase)
        Me.ClearGrid()
        Me.DataBindings.Clear()
        Me.ResetLayouts()
        Initialize(globalObj)

    End Sub
    ''' <summary>
    ''' Initialize internal objects.
    ''' </summary>
    ''' <param name="globalObj"></param>
    ''' <remarks></remarks>
    Public Overridable Sub Initialize(ByRef globalObj As Bradford.ObjectLibrary.IGlobalObjectBase)
        Try
            _GlobalObject = globalObj
            _DB = New DatabaseObject.BSCUltraDataGridDB(globalObj)
            _UniqueContainerName = GenerateUniqueContainerName()

            BuildContextMenu()
            BuildSmartTagMenuLeft()
        Catch ex As Exception
            globalObj.HandleEx(ex, Me.UniqueContainerName & " Initializing.")
        End Try

    End Sub

    ''' <summary>
    ''' Set the DataSource using a IBindingList
    ''' </summary>
    ''' <param name="dataSource"></param>
    ''' <remarks></remarks>
    Public Overloads Sub SetDataSource(Of T As IBindingList)(ByVal dataSource As T, Optional ByVal dataMember As String = "")
        Try
            Dim oArg As New Bradford.Common.CancelEventArg

            OnBeforeLoadGrid(oArg)

            If oArg.Cancel Then
                Exit Sub
            End If

            Me.SetDataBinding(dataSource, dataMember)

            OnAfterLoadGrid()
        Catch ex As Exception
            _ExManager.ShowForm(ex, "Setting Grid Data Source using IBindingList")
        End Try

    End Sub
    ''' <summary>
    ''' Set the DataSource using a DataSet
    ''' </summary>
    ''' <param name="dataSource"></param>
    ''' <remarks></remarks>
    Public Overloads Sub SetDataSource(ByVal dataSource As DataSet)
        Try
            Dim oArg As New Bradford.Common.CancelEventArg

            OnBeforeLoadGrid(oArg)

            If oArg.Cancel Then
                Exit Sub
            End If

            MyBase.SetDataBinding(dataSource, String.Empty)

            OnAfterLoadGrid()

        Catch ex As Exception
            _ExManager.ShowForm(ex, "Setting Grid Data Source using DataSet")
        End Try
    End Sub

    ''' <summary>
    ''' Set the DataSource using a DataTable
    ''' </summary>
    ''' <param name="dataSource"></param>
    ''' <remarks></remarks>
    Public Overloads Sub SetDataSource(ByVal dataSource As DataTable)
        Try
            Dim oArg As New Bradford.Common.CancelEventArg

            OnBeforeLoadGrid(oArg)

            If oArg.Cancel Then
                Exit Sub
            End If

            MyBase.SetDataBinding(dataSource, String.Empty)


            OnAfterLoadGrid()

        Catch ex As Exception
            _ExManager.ShowForm(ex, "Setting Grid Data Source using DataTable")
        End Try
    End Sub

    ''' <summary>
    ''' Clears the selected rows.
    ''' </summary>
    ''' <developer>Pete Sral</developer>
    Public Sub ClearSelectedRows()
        Try
            SelectedRows.Clear()
            Selected.Rows.Clear()
            Me.ReInitializeRows()
        Catch ex As Exception
            HandleEx(ex, "Clearing selected rows.")
        End Try

    End Sub
    ''' <summary>
    ''' Clear Grid
    ''' </summary>
    ''' <remarks></remarks>
    Public Overridable Sub ClearGrid()
        Try

            HideMessages()

            If Me.DataSource IsNot Nothing Then
                Me.DataSource = Nothing
            End If

            SelectedRows.Clear()
        Catch ex As Exception
            Throw New Exception("Clearing Grid", ex)
        End Try

    End Sub
    ''' <summary>
    ''' Called execute load grid..
    ''' </summary>
    ''' <developer>Pete Sral</developer>
    Public Overridable Sub OnExecuteLoadGrid()
        SelectedRows.Clear()
        RaiseEvent ExecuteLoadGrid()
    End Sub
    ''' <summary>
    ''' Called before load grid.
    ''' </summary>
    ''' <param name="e">The e.</param>
    ''' <developer>Pete Sral</developer>
    Public Overridable Sub OnBeforeLoadGrid(ByVal e As CancelEventArg)
        HideMessages()
        RaiseEvent BeforeLoadGrid(Me, e)
    End Sub

    ''' <summary>
    ''' Called after load grid.
    ''' </summary>
    ''' <developer>Pete Sral</developer>
    Public Overridable Sub OnAfterLoadGrid()

        SelectedRows.Clear()
        HandleLayout()

        RaiseEvent AfterLoadGrid(Me, New EventArgs)

    End Sub

    ''' <summary>
    ''' Test if column exists.
    ''' </summary>
    ''' <param name="columnName"></param>
    ''' <param name="band"></param>
    ''' <returns></returns>
    ''' <remarks>Same as <code>DisplayLayout.Bands(band).Columns.Exists(key)</code> </remarks>
    Public Overridable Function ColumnExists(ByVal columnName As String, Optional ByVal band As Integer = 0) As Boolean
        Try
            If String.IsNullOrEmpty(columnName) Then
                Return False
            End If
            Return Me.DisplayLayout.Bands(band).Columns.Exists(columnName)
        Catch ex As Exception
            Return False
        End Try
    End Function
    ''' <summary>
    ''' Hide column by index
    ''' </summary>
    ''' <param name="index"></param>
    ''' <remarks></remarks>
    Public Sub HideColumn(ByVal index As Integer, Optional ByVal value As Boolean = True, Optional ByVal layout As Integer = 0, Optional ByVal band As Integer = 0)
        Try
            Dim oColItem As New HiddenColumnItem

            With oColItem
                .Band = band
                .Index = index
                .Layout = layout
            End With

            If Not _HiddenColumns.ContainsKey(index.ToString) Then
                _HiddenColumns.Add(index.ToString, oColItem)
            End If


            If Me.Layouts.Count <> 0 Then
                Me.Layouts(layout).Bands(band).Columns(index).Hidden = value
            Else
                DisplayLayout.Bands(band).Columns(index).Hidden = value
            End If


        Catch ex As Exception
            _ExManager.ShowForm(ex, "Hiding Column Index: " & index.ToString)
        End Try
    End Sub
    ''' <summary>
    ''' Hide column by column Name
    ''' </summary>
    ''' <param name="columnName"></param>
    ''' <remarks></remarks>
    Public Sub HideColumn(ByVal columnName As String, Optional ByVal value As Boolean = True, Optional ByVal layout As Integer = 0, Optional ByVal band As Integer = 0)
        Try

            If Not ColumnExists(columnName, band) Then
                Dim oColItem As New HiddenColumnItem

                With oColItem
                    .Band = band
                    .ColumnName = columnName
                    .Layout = layout
                End With

                If Not _HiddenColumns.ContainsKey(oColItem.ColumnName) Then
                    _HiddenColumns.Add(oColItem.ColumnName, oColItem)
                End If


                Exit Sub
            End If


            If Layouts.Count <> 0 Then
                Layouts(layout).Bands(band).Columns(columnName).Hidden = value
                'oColItem.Index = Layouts(layout).Bands(band).Columns(columnName).Index
            Else
                Me.DisplayLayout.Bands(band).Columns(columnName).Hidden = value
                'oColItem.Index = Me.DisplayLayout.Bands(band).Columns(columnName).Index
            End If



        Catch ex As Exception
            _ExManager.ShowForm(ex, "Hiding Column Name: " & columnName)
        End Try
    End Sub
    ''' <summary>
    ''' Set the columns header cell value (caption).
    ''' </summary>
    ''' <param name="columnName"></param>
    ''' <param name="caption"></param>
    ''' <remarks></remarks>
    Public Sub SetCaption(ByVal columnName As String, ByVal caption As String, Optional ByVal layout As Integer = 0, Optional ByVal band As Integer = 0)
        Try
            If columnName = String.Empty OrElse caption = String.Empty Then
                Exit Sub
            End If

            If Not ColumnExists(columnName, band) Then
                Exit Sub
            End If

            If Layouts.Count <> 0 Then
                Layouts(layout).Bands(band).Columns(columnName).Header.Caption = caption
            Else
                DisplayLayout.Bands(band).Columns(columnName).Header.Caption = caption
            End If

        Catch ex As Exception
            _ExManager.ShowForm(ex, "Setting Column : " & columnName & " Header Caption: " & caption)
        End Try
    End Sub
    ''' <summary>
    ''' Set the column position within the band.
    ''' </summary>
    ''' <param name="columnName"></param>
    ''' <param name="position"></param>
    ''' <param name="band"></param>
    ''' <remarks></remarks>
    Public Sub SetColumnPosition(ByVal columnName As String, ByVal position As Integer, Optional ByVal band As Integer = 0)
        Try

            If Not ColumnExists(columnName, band) Then
                Exit Sub
            End If

            DisplayLayout.Bands(band).Columns(columnName).Header.VisiblePosition = position

        Catch ex As Exception
            _ExManager.ShowForm(ex, "Setting Column : " & columnName & " Position: " & position.ToString & " Band: " & band.ToString)
        End Try
    End Sub
    ''' <summary>
    ''' Set format string to column.
    ''' </summary>
    ''' <param name="columnName"></param>
    ''' <param name="format"></param>
    ''' <remarks></remarks>
    Public Sub SetColumnFormat(ByVal columnName As String, ByVal format As String, Optional ByVal band As Integer = 0)
        Try

            If Not ColumnExists(columnName, band) Then
                Exit Sub
            End If

            DisplayLayout.Bands(band).Columns(columnName).Format = format

        Catch ex As Exception
            _ExManager.ShowForm(ex, "Setting Column Format : " & columnName & " Format: " & format)
        End Try
    End Sub
    ''' <summary>
    ''' Retrieve UltraGridColumn based on column name.
    ''' </summary>
    ''' <param name="columnName"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Function GetColumn(ByVal columnName As String) As Infragistics.Win.UltraWinGrid.UltraGridColumn
        Try
            If Me.DataSource IsNot Nothing Then
                Return Me.Rows.Band.Columns(columnName)
            Else
                Return Nothing
            End If

        Catch ex As Exception
            _ExManager.ShowForm(ex, "Getting column: " & columnName)
            Return Nothing
        End Try
    End Function
    ''' <summary>
    ''' Select the row for the column name and matching value.
    ''' </summary>
    ''' <param name="columnName"></param>
    ''' <param name="value"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Function SelectRow(ByVal columnName As String, ByVal value As Object) As Boolean
        Dim oCol As Infragistics.Win.UltraWinGrid.UltraGridColumn = GetColumn(columnName)

        If oCol Is Nothing Then
            Return False
        End If

        Try
            Dim rows As UltraGridRow() = Me.Rows.GetAllNonGroupByRows
            For Each row As UltraGridRow In rows

                If String.Compare(GetValue(row.GetCellValue(columnName)), GetValue(value), True) = 0 Then
                    row.Activate()
                    row.Selected = True

                    Return True
                ElseIf row.ChildBands IsNot Nothing AndAlso row.ChildBands.HasChildRows Then
                    For i As Integer = 0 To row.ChildBands.Count - 1
                        If SelectRow(row.ChildBands(i).Rows, columnName, value) Then
                            Return True
                        End If
                    Next
                End If
            Next

        Catch ex As Exception
            _ExManager.ShowForm(ex, "Selecting Row for Column: " & columnName & " = Value: " & value.ToString)

        End Try
    End Function

    ''' <summary>
    ''' Selects the rows.
    ''' </summary>
    ''' <param name="rows">The rows.</param>
    ''' <param name="columnName">Name of the column.</param>
    ''' <param name="value">The value.</param>
    ''' <returns></returns>
    ''' <developer>Pete Sral</developer>
    Public Function SelectRow(ByVal rows As RowsCollection, ByVal columnName As String, ByVal value As Object) As Boolean
        Dim oCol As Infragistics.Win.UltraWinGrid.UltraGridColumn = GetColumn(columnName)

        If oCol Is Nothing Then
            Return False
        End If

        Try
            Dim oRows As UltraGridRow() = rows.GetAllNonGroupByRows
            For Each row As UltraGridRow In oRows

                If String.Compare(GetValue(row.GetCellValue(columnName)), GetValue(value), True) = 0 Then
                    row.Activate()
                    row.Selected = True
                    Return True
                End If
            Next

        Catch ex As Exception
            _ExManager.ShowForm(ex, "Selecting Row for Column: " & columnName & " = Value: " & value.ToString)

        End Try


    End Function


    ''' <summary>
    ''' Recurs through all parent control to build a control container path.
    ''' </summary>
    ''' <param name="ctrl"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Function GetControlPath(ByVal ctrl As Control) As String

        If ctrl.Parent Is Nothing Then
            Return _UniqueContainerName
        Else
            If Not Me.Equals(ctrl) And Not String.IsNullOrEmpty(ctrl.Name) Then
                _UniqueContainerName = ctrl.Name & "." & _UniqueContainerName
            End If

            Return GetControlPath(ctrl.Parent)
        End If



    End Function
    ''' <summary>
    ''' Called when display progress.
    ''' </summary>
    ''' <param name="value">The value.</param>
    ''' <param name="total">The total.</param>
    ''' <developer>Pete Sral</developer>
    Public Overridable Sub OnDisplayProgress(ByVal value As Integer, ByVal total As Integer)

        If value < total Then
            'RaiseEvent DisplayProgress(value, total)

            If ProgressBar IsNot Nothing Then

                ProgressBar.Maximum = total
                ProgressBar.Minimum = 1
                ProgressBar.Step = 1
                ProgressBar.Value = value
                ProgressBar.Enabled = True
                ProgressBar.Invalidate()
            End If


            'Debug.Print(value, total)
            'Bradford.Common.WinForms.ProgressDlg.Update(value, total)

        Else
            'RaiseEvent DisplayProgress(0, 0)

            If ProgressBar IsNot Nothing Then
                ProgressBar.Maximum = 0
                ProgressBar.Value = 0

            End If
        End If

        'Application.DoEvents()
    End Sub

    ''' <summary>
    ''' Called when set message.
    ''' </summary>
    ''' <param name="message">The message.</param>
    ''' <developer>Pete Sral</developer>
    Public Overridable Sub OnSetMessage(ByVal message As String)
        OnSetMessage(message, String.Empty)
    End Sub
    ''' <summary>
    ''' Called when set message.
    ''' </summary>
    ''' <param name="message">The message.</param>
    ''' <param name="imageID">The image ID.</param>
    ''' <developer>Pete Sral</developer>
    Public Overridable Sub OnSetMessage(ByVal message As String, ByVal imageID As String)
        If IsInitialized Then
            If Not GlobalObject.TopMessage Is Nothing Then
                GlobalObject.TopMessage.Add(message, imageID)
            End If
        End If
    End Sub




    ''' <summary>
    ''' Browse grid layout.
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="layout "></param>
    ''' <remarks></remarks>
    Public Overridable Sub InitializeLayoutBrowse(ByVal sender As Object, ByVal layout As UltraGridLayout)
        'only show the column header once at the top
        'e.Layout.Override.HeaderPlacement = Infragistics.Win.UltraWinGrid.HeaderPlacement.OncePerRowIsland
        'e.Layout.Bands(1).ColHeadersVisible = False
        'e.Layout.Override.AllowColSizing = Infragistics.Win.UltraWinGrid.AllowColSizing.Synchronized

        If Me.Rows.Count > AutoResizeAllRowCount Then
            layout.PerformAutoResizeColumns(False, Infragistics.Win.UltraWinGrid.PerformAutoSizeType.VisibleRows, True)
        Else
            layout.PerformAutoResizeColumns(False, Infragistics.Win.UltraWinGrid.PerformAutoSizeType.AllRowsInBand, True)
        End If



        With layout.Override
            .ExpansionIndicator = Infragistics.Win.UltraWinGrid.ShowExpansionIndicator.CheckOnDisplay

            'handle how row is selected
            If GridLayoutStyle = BSCUltraDataGridStyle.BrowseSingleSelect Then
                .SelectTypeCell = Infragistics.Win.UltraWinGrid.SelectType.None
                .SelectTypeRow = Infragistics.Win.UltraWinGrid.SelectType.Single
                .CellClickAction = Infragistics.Win.UltraWinGrid.CellClickAction.RowSelect
                .SelectTypeGroupByRow = Infragistics.Win.UltraWinGrid.SelectType.Single
            ElseIf GridLayoutStyle = BSCUltraDataGridStyle.BrowseMultiSelect Then
                .SelectTypeCell = Infragistics.Win.UltraWinGrid.SelectType.None
                .SelectTypeRow = Infragistics.Win.UltraWinGrid.SelectType.ExtendedAutoDrag
                .CellClickAction = Infragistics.Win.UltraWinGrid.CellClickAction.RowSelect
                .SelectTypeGroupByRow = Infragistics.Win.UltraWinGrid.SelectType.ExtendedAutoDrag
            End If


            .AllowAddNew = Infragistics.Win.UltraWinGrid.AllowAddNew.No
            .AllowDelete = Infragistics.Win.DefaultableBoolean.False
            .AllowUpdate = Infragistics.Win.DefaultableBoolean.False

            'turn filtering on
            .FilterUIType = FilterUIType.FilterRow
            .AllowRowFiltering = Infragistics.Win.DefaultableBoolean.True
            .RowFilterMode = RowFilterMode.AllRowsInBand
            .RowFilterAction = RowFilterAction.HideFilteredOutRows
            .FilterComparisonType = FilterComparisonType.CaseInsensitive


            .FixedHeaderIndicator = Infragistics.Win.UltraWinGrid.FixedHeaderIndicator.InSwapDropDown

            .AllowColMoving = Infragistics.Win.UltraWinGrid.AllowColMoving.WithinBand
            .AllowColSizing = Infragistics.Win.UltraWinGrid.AllowColSizing.Synchronized
            .SortComparisonType = Infragistics.Win.UltraWinGrid.SortComparisonType.CaseInsensitive
            .HeaderClickAction = Infragistics.Win.UltraWinGrid.HeaderClickAction.SortMulti


            'show row number in selector column
            .RowSelectorNumberStyle = Infragistics.Win.UltraWinGrid.RowSelectorNumberStyle.RowIndex



        End With

    End Sub

    Public Overridable Sub InitializeLayoutStandard(ByVal sender As Object, ByVal layout As UltraGridLayout)
        'only show the column header once at the top
        'e.Layout.Override.HeaderPlacement = Infragistics.Win.UltraWinGrid.HeaderPlacement.OncePerRowIsland
        'e.Layout.Bands(1).ColHeadersVisible = False
        'e.Layout.Override.AllowColSizing = Infragistics.Win.UltraWinGrid.AllowColSizing.Synchronized

        If Me.Rows.Count > AutoResizeAllRowCount Then
            layout.PerformAutoResizeColumns(False, Infragistics.Win.UltraWinGrid.PerformAutoSizeType.VisibleRows, True)
        Else
            layout.PerformAutoResizeColumns(False, Infragistics.Win.UltraWinGrid.PerformAutoSizeType.AllRowsInBand, True)
        End If


        With layout.Override
            .ExpansionIndicator = Infragistics.Win.UltraWinGrid.ShowExpansionIndicator.CheckOnDisplay

            'turn filtering on
            .FilterUIType = FilterUIType.FilterRow
            .AllowRowFiltering = Infragistics.Win.DefaultableBoolean.True
            .RowFilterMode = RowFilterMode.AllRowsInBand
            .RowFilterAction = RowFilterAction.HideFilteredOutRows
            .FilterComparisonType = FilterComparisonType.CaseInsensitive


            .FixedHeaderIndicator = Infragistics.Win.UltraWinGrid.FixedHeaderIndicator.InSwapDropDown

            .AllowColMoving = Infragistics.Win.UltraWinGrid.AllowColMoving.WithinBand
            .AllowColSizing = Infragistics.Win.UltraWinGrid.AllowColSizing.Synchronized
            .SortComparisonType = Infragistics.Win.UltraWinGrid.SortComparisonType.CaseInsensitive
            .HeaderClickAction = Infragistics.Win.UltraWinGrid.HeaderClickAction.SortMulti


            'show row number in selector column
            .RowSelectorNumberStyle = Infragistics.Win.UltraWinGrid.RowSelectorNumberStyle.RowIndex



        End With


    End Sub
#End Region


#Region "Context Menu"

    Private _GridContextMenu As ContextMenu
    ''' <summary>
    ''' Custom Context Menu
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Property GridContextMenu() As ContextMenu
        Get

            Return _GridContextMenu
        End Get
        Set(ByVal Value As ContextMenu)
            _GridContextMenu = Value
        End Set
    End Property

    ''' <summary>
    ''' Custom Context Menu for Grid.
    ''' </summary>
    ''' <remarks></remarks>
    Public Overridable Sub BuildContextMenu()

        If GridContextMenu IsNot Nothing Then
            Exit Sub
        Else
            GridContextMenu = New ContextMenu
        End If

        'Try
        '    With GridContextMenu
        '        Dim oSep As New MenuItem("-")

        '        'Export
        '        .MenuItems.Add("Grid Menu")
        '        .MenuItems.Add("-")
        '        '.MenuItems.Add("Columns", New EventHandler(AddressOf EventHandlerColumns))
        '        .MenuItems.Add("Export", New EventHandler(AddressOf EventHandlerExport))


        '        oSep = Nothing

        '    End With
        'Catch ex As Exception
        '    Throw
        'End Try
    End Sub
#End Region

#Region "Smart Tag Menu"
    Private _SmartTagMenuLayout As New ToolStripMenuItem("Layout")

    ''' <summary>
    ''' Gets the smart tag menu layout.
    ''' </summary>
    ''' <value>The smart tag menu layout.</value>
    ''' <developer>Pete Sral</developer>
    Public Overridable ReadOnly Property SmartTagMenuLayout() As ToolStripMenuItem
        Get
            Return _SmartTagMenuLayout
        End Get
    End Property

    ''' <summary>
    ''' Gets the smart tag menu layout save.
    ''' </summary>
    ''' <value>The smart tag menu layout save.</value>
    ''' <developer>Pete Sral</developer>
    Public Overridable ReadOnly Property SmartTagMenuLayoutSave() As ToolStripMenuItem
        Get
            Return _SmartTagMenuLayoutSave
        End Get
    End Property

    ''' <summary>
    ''' Gets the smart tag menu layout load.
    ''' </summary>
    ''' <value>The smart tag menu layout load.</value>
    ''' <developer>Pete Sral</developer>
    Public Overridable ReadOnly Property SmartTagMenuLayoutLoad() As ToolStripMenuItem
        Get
            Return _SmartTagMenuLayoutLoad
        End Get
    End Property

    ''' <summary>
    ''' Gets the smart tag menu layout delete.
    ''' </summary>
    ''' <value>The smart tag menu layout delete.</value>
    ''' <developer>Pete Sral</developer>
    Public Overridable ReadOnly Property SmartTagMenuLayoutDelete() As ToolStripMenuItem
        Get
            Return _SmartTagMenuLayoutDelete
        End Get
    End Property


    ''' <summary>
    ''' Gets the smart tag menu layout reset.
    ''' </summary>
    ''' <value>The smart tag menu layout reset.</value>
    ''' <developer>Pete Sral</developer>
    Public Overridable ReadOnly Property SmartTagMenuLayoutReset() As ToolStripMenuItem
        Get
            Return _SmartTagMenuLayoutReset
        End Get
    End Property

    ''' <summary>
    ''' Gets the smart tag menu restore sort.
    ''' </summary>
    ''' <value>The smart tag menu restore sort.</value>
    ''' <developer>Pete Sral</developer>
    Public Overridable ReadOnly Property SmartTagMenuRestoreSort() As ToolStripMenuItem
        Get
            Return _SmartTagMenuRestoreSort
        End Get
    End Property

    ''' <summary>
    ''' Gets the smart tag menu export excel.
    ''' </summary>
    ''' <value>The smart tag menu export excel.</value>
    ''' <developer>Pete Sral</developer>
    Public Overridable ReadOnly Property SmartTagMenuExportExcel() As ToolStripMenuItem
        Get
            Return _SmartTagMenuExportExcel
        End Get
    End Property

    ''' <summary>
    ''' Gets the smart tag menu column chooser.
    ''' </summary>
    ''' <value>The smart tag menu column chooser.</value>
    ''' <developer>Pete Sral</developer>
    Public Overridable ReadOnly Property SmartTagMenuColumnChooser() As ToolStripMenuItem
        Get
            Return _SmartTagMenuColumnChooser
        End Get
    End Property


    'save layout
    Private WithEvents _SmartTagMenuLayoutSave As New ToolStripMenuItem("Save")
    Private Sub _SmartTagMenuLayoutSave_Click(ByVal sender As Object, ByVal e As System.EventArgs) Handles _SmartTagMenuLayoutSave.Click
        SaveLayout()
    End Sub

    'load layout
    Private WithEvents _SmartTagMenuLayoutLoad As New ToolStripMenuItem("Load")
    Private Sub _SmartTagMenuLayoutLoad_Click(ByVal sender As Object, ByVal e As System.EventArgs) Handles _SmartTagMenuLayoutLoad.Click
        LoadLayout()
    End Sub

    'reset layout
    Private WithEvents _SmartTagMenuLayoutReset As New ToolStripMenuItem("Reset")
    Private Sub _SmartTagMenuLayoutReset_Click(ByVal sender As Object, ByVal e As System.EventArgs) Handles _SmartTagMenuLayoutReset.Click
        ResetLayout()
    End Sub

    'remove layout from database 
    Private WithEvents _SmartTagMenuLayoutDelete As New ToolStripMenuItem("Delete")
    Private Sub _SmartTagMenuLayoutDelete_Click(ByVal sender As Object, ByVal e As System.EventArgs) Handles _SmartTagMenuLayoutDelete.Click
        DeleteLayout()
    End Sub
    'restore default sort
    Private WithEvents _SmartTagMenuRestoreSort As New ToolStripMenuItem("Restore Default Sort")
    Private Sub _SmartTagMenuRestoreSort_Click(ByVal sender As Object, ByVal e As System.EventArgs) Handles _SmartTagMenuRestoreSort.Click
        RestoreDefaultSort()
    End Sub

    Private WithEvents _SmartTagMenuExportExcel As New ToolStripMenuItem("Excel...")
    Private Sub _SmartTagMenuExportExcel_Click(ByVal sender As Object, ByVal e As System.EventArgs) Handles _SmartTagMenuExportExcel.Click
        ExportExcel()
    End Sub

    Private WithEvents _SmartTagMenuColumnChooser As New ToolStripMenuItem("Column Chooser...")
    Private Sub _SmartTagMenuColumnChooser_Click(ByVal sender As Object, ByVal e As System.EventArgs) Handles _SmartTagMenuColumnChooser.Click
        ShowColumnChooser()
    End Sub

    Private WithEvents _SmartTagMenuExport As New ToolStripMenuItem("Export")
    ''' <summary>
    ''' Gets the smart tag menu export.
    ''' </summary>
    ''' <value>The smart tag menu export.</value>
    ''' <developer>Pete Sral</developer>
    Public ReadOnly Property SmartTagMenuExport As ToolStripMenuItem
        Get
            Return _SmartTagMenuExport
        End Get
    End Property

    ''' <summary>
    ''' Smarts the tag menu security.
    ''' </summary>
    ''' <developer>Pete Sral</developer>
    Public Overridable Sub SmartTagMenuState()
        Try
            SmartTagMenuRestoreSort.Enabled = Me.DataSource IsNot Nothing AndAlso IsInitialized
            SmartTagMenuLayout.Enabled = Me.Rows IsNot Nothing AndAlso Me.Rows.Count <> 0 AndAlso IsInitialized
            SmartTagMenuExport.Enabled = Me.Rows IsNot Nothing AndAlso Me.Rows.Count <> 0
            SmartTagMenuExportExcel.Enabled = SmartTagMenuExport.Enabled

            SmartTagMenuColumnChooser.Enabled = Me.DataSource IsNot Nothing AndAlso Me.Rows IsNot Nothing AndAlso Me.Rows.Count <> 0

            If SmartTagMenuLayout.Enabled Then
                If DB IsNot Nothing Then
                    SmartTagMenuLayoutLoad.Enabled = IsInitialized AndAlso DB.SelectLayoutCount(Me.UniqueContainerName, Me.Name) <> 0
                End If

                SmartTagMenuLayoutDelete.Enabled = HasSavedLayout
                SmartTagMenuLayoutReset.Enabled = HasSavedLayout
                SmartTagMenuLayoutSave.Enabled = True

            Else
                SmartTagMenuLayoutLoad.Enabled = False
                SmartTagMenuLayoutReset.Enabled = False
                SmartTagMenuLayoutSave.Enabled = False
                SmartTagMenuLayoutDelete.Enabled = False
            End If



        Catch ex As Exception
            HandleEx(ex, "Setting Smart Tag Menu Security: " & Me.UniqueContainerName)
        End Try
    End Sub
#End Region

#Region "Smart Tag"
    Private _SmartTagMenuLeft As ContextMenuStrip
    ''' <summary>
    ''' Left click smart tag menu
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Property SmartTagMenuLeft() As ContextMenuStrip
        Get
            Return _SmartTagMenuLeft
        End Get
        Set(ByVal value As ContextMenuStrip)
            _SmartTagMenuLeft = value
        End Set
    End Property


    Private _SmartTagButton As PictureBox
    ''' <summary>
    ''' Gets or sets the smart tag button.
    ''' </summary>
    ''' <value>The smart tag button.</value>
    ''' <developer>Pete Sral</developer>
    Public Property SmartTagButton() As PictureBox
        Get
            Return _SmartTagButton
        End Get
        Set(ByVal value As PictureBox)
            _SmartTagButton = value
        End Set
    End Property

    ''' <summary>
    ''' Builds the smart tag menu left.
    ''' </summary>
    ''' <developer>Pete Sral</developer>
    Public Overridable Sub BuildSmartTagMenuLeft()
        Try
            If SmartTagMenuLeft Is Nothing Then
                _SmartTagMenuLeft = New ContextMenuStrip
            Else
                'already setup
                Exit Sub
            End If

            'Save Layouts

            SmartTagMenuLayout.Enabled = False

            With SmartTagMenuLayout
                .Enabled = False
                .Name = "SmartTagLayout"
                .Image = My.Resources.options

                'save
                SmartTagMenuLayoutSave.Image = My.Resources.save
                SmartTagMenuLayoutSave.Name = "SmartTagMenuLayoutSave"
                .DropDownItems.Add(SmartTagMenuLayoutSave)

                'load
                SmartTagMenuLayoutLoad.Image = My.Resources.open
                SmartTagMenuLayoutLoad.Name = "SmartTagLayoutLoad"
                .DropDownItems.Add(SmartTagMenuLayoutLoad)

                'delete
                SmartTagMenuLayoutDelete.Image = My.Resources.Delete
                SmartTagMenuLayoutDelete.Name = "SmartTagLayoutDelete"
                .DropDownItems.Add(SmartTagMenuLayoutDelete)

                .DropDownItems.Add("-")

                'reset
                SmartTagMenuLayoutReset.Image = My.Resources.refresh
                SmartTagMenuLayoutReset.Name = "SmartTagMenuLayoutReset"
                .DropDownItems.Add(SmartTagMenuLayoutReset)


            End With
            SmartTagMenuLeft.Items.Add(SmartTagMenuLayout)

            SmartTagMenuLeft.Items.Add("-")

            'Restore Store
            SmartTagMenuRestoreSort.Name = "SmartTagMenuRestore"
            SmartTagMenuLeft.Items.Add(SmartTagMenuRestoreSort)

            SmartTagMenuLeft.Items.Add("-")


            'Export
            SmartTagMenuExport.Image = My.Resources.table_export
            SmartTagMenuLeft.Items.Add(SmartTagMenuExport)

            With SmartTagMenuExport
                .Name = "SmartTagExport"
                .Text = "Export"
                'Excel
                SmartTagMenuExportExcel.Image = My.Resources.document_excel
                SmartTagMenuExportExcel.Name = "SmartTagExportExcel"

                SmartTagMenuExport.DropDownItems.Add(SmartTagMenuExportExcel)

            End With

            With SmartTagMenuColumnChooser
                .Name = "SmartTagColumnChooser"
                .Image = My.Resources.grid_column_width

                SmartTagMenuLeft.Items.Add(SmartTagMenuColumnChooser)
            End With

        Catch ex As Exception
            If DesignMode Then
                Exit Sub
            End If

        End Try
    End Sub
    ''' <summary>
    ''' Handles the mouse up.
    ''' </summary>
    ''' <param name="sender">The sender.</param>
    ''' <param name="e">The <see cref="System.Windows.Forms.MouseEventArgs" /> instance containing the event data.</param>
    ''' <developer>Pete Sral</developer>
    Public Overridable Sub HandleMouseUp(ByVal sender As Object, ByVal e As MouseEventArgs)
        Select Case e.Button
            Case Windows.Forms.MouseButtons.Left
                If SmartTagMenuLeft Is Nothing Then
                    BuildSmartTagMenuLeft()
                End If
                SmartTagMenuState()
                SmartTagMenuLeft.Show(SmartTagButton, e.X, e.Y)

            Case Windows.Forms.MouseButtons.Right


        End Select
    End Sub

    ''' <summary>
    ''' Shows the smart tag button.
    ''' </summary>
    ''' <developer>Pete Sral</developer>
    Public Overridable Sub ShowSmartTagButton()
        If SmartTagButton Is Nothing Then
            _SmartTagButton = New PictureBox

            With SmartTagButton
                AddHandler SmartTagButton.MouseUp, AddressOf HandleMouseUp

                .Image = My.Resources.ChevronArrowDown_Small
                .SizeMode = PictureBoxSizeMode.AutoSize
                .BorderStyle = BorderStyle.None
                '.Location = New Point(Me.Left, Me.Top)
                Me.Controls.Add(SmartTagButton)

            End With
        End If

        With SmartTagButton
            'position
            .Left = 1  'Me.Left + Me.Width - _SmartTag.Width
            .Top = 1

            .Invalidate()
            .Visible = True
            .BringToFront()
        End With




    End Sub

    ''' <summary>
    ''' Hides the smart tag button.
    ''' </summary>
    ''' <developer>Pete Sral</developer>
    Public Overridable Sub HideSmartTagButton()
        If SmartTagButton IsNot Nothing Then
            SmartTagButton.Visible = False
        End If
    End Sub
#End Region

#Region "Layout"
    ''' <summary>
    '''  (READ ONLY) Gets the current display layout.
    ''' </summary>
    ''' <value>The current layout.</value>
    ''' <developer>Pete Sral</developer>
    <Browsable(False)> _
    Public Property CurrentLayout() As UltraGridLayout
        Get
            Return Me.DisplayLayout
        End Get
        Set(ByVal value As UltraGridLayout)
            'backward compatibility for existing forms using this control.
        End Set
    End Property

    Private _SavedLayout As UltraGridLayout
    ''' <summary>
    ''' Gets or sets the saved layout.
    ''' </summary>
    ''' <value>The saved layout.</value>
    ''' <developer>Pete Sral</developer>
    <Browsable(False)> _
    Public Property SavedLayout() As UltraGridLayout
        Get
            Return _SavedLayout
        End Get
        Set(ByVal value As UltraGridLayout)
            _SavedLayout = value
        End Set
    End Property

    Private _DefaultLayout As UltraGridLayout
    ''' <summary>
    ''' Gets or sets the default layout.
    ''' </summary>
    ''' <value>The default layout.</value>
    ''' <developer>Pete Sral</developer>
    <Browsable(False)> _
    Public Property DefaultLayout() As UltraGridLayout
        Get
            Return _DefaultLayout
        End Get
        Set(ByVal value As UltraGridLayout)
            _DefaultLayout = value
        End Set
    End Property


    ''' <summary>
    ''' Restore default sort
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Public Overridable Sub HandleRestoreDefaultSort(ByVal sender As Object, ByVal e As EventArgs)
        RestoreDefaultSort()
    End Sub
    ''' <summary>
    ''' Reload grid to restore default sort.
    ''' </summary>
    ''' <remarks></remarks>
    Public Overridable Sub RestoreDefaultSort()
        OnExecuteLoadGrid()
    End Sub
    ''' <summary>
    ''' Reset layout to the default layout.
    ''' </summary>
    ''' <remarks></remarks>
    Public Overridable Sub ResetLayout()
        RestoreDefaultSort()

        If DefaultLayout Is Nothing Then
            InitializeCurrentLayout(Me, CurrentLayout)
            _DefaultLayout = CurrentLayout
        End If
        ResetLayout(_DefaultLayout)

    End Sub
    ''' <summary>
    ''' Apply layout to grid.
    ''' </summary>
    ''' <param name="layout"></param>
    ''' <remarks></remarks>
    Public Sub ResetLayout(ByVal layout As UltraGridLayout)
        If layout IsNot Nothing Then
            Me.DisplayLayout.Load(layout.Clone, PropertyCategories.All)
        End If
    End Sub


    ''' <summary>
    ''' Saves the layout.
    ''' </summary>
    ''' <developer>Pete Sral</developer>
    Public Overridable Sub SaveLayout()
        'http://support.infragistics.com/Support/KnowledgeBaseArticle.aspx?ArticleID=1608
        Try

            If Not IsInitialized Then
                Exit Sub
            End If

            'Step 1 - Save the layout to a stream
            Dim MS As New IO.MemoryStream()
            Me.DisplayLayout.Save(MS, Infragistics.Win.UltraWinGrid.PropertyCategories.All)

            'Step 2 - Store the stream in a Byte array
            Dim LayoutArray() As Byte
            LayoutArray = MS.ToArray()

            'Step 3 - Once the Layout is stored in a stream, this value can be written to the appropriate field in a DataTable. 
            'GridLayoutsTable.Rows.Add(New Object() {Nothing, "MyGridLayout", LayoutArray})
            Dim oGridLayoutTable As DataTable = DB.SelectLayoutDataTable(Me.UniqueContainerName, Me.Name)

            If oGridLayoutTable Is Nothing OrElse oGridLayoutTable.Rows.Count = 0 Then
                oGridLayoutTable.Rows.Add(New Object() {GlobalObject.ApplicationInfo.ID, GlobalObject.Security.UserID, Me.UniqueContainerName, Me.Name, "", LayoutArray, GlobalObject.Now, NewGuid()})
            Else
                oGridLayoutTable.Rows(0).Item("Layout") = LayoutArray
                oGridLayoutTable.Rows(0).Item("Layout_TS") = GlobalObject.Now
            End If

            If DB.SaveLayoutDataTable(oGridLayoutTable) Then
                GlobalObject.TopMessage.Add("Grid layout saved.")
            End If

            _SavedLayout = Me.DisplayLayout.Clone


        Catch ex As Exception
            _ExManager.ShowForm(ex, "Saving layout for: " & Me.UniqueContainerName & Me.Name)
        End Try
    End Sub


    ''' <summary>
    ''' Gets a value indicating whether this instance has a saved layout.
    ''' </summary>
    ''' <value>
    ''' <c>true</c> if this instance has saved layout; otherwise, <c>false</c>.
    ''' </value>
    ''' <developer>Pete Sral</developer>
    ReadOnly Property HasSavedLayout As Boolean
        Get
            Return SavedLayout IsNot Nothing
        End Get
    End Property
    ''' <summary>
    ''' Loads the layout.
    ''' </summary>
    ''' <developer>Pete Sral</developer>
    Public Overridable Function LoadLayout() As Boolean
        'http://support.infragistics.com/Support/KnowledgeBaseArticle.aspx?ArticleID=1608
        Try
            If _DefaultLayout Is Nothing Then
                InitializeCurrentLayout(Me, Me.DisplayLayout)
                _DefaultLayout = Me.DisplayLayout.Clone
            End If

            If Not IsInitialized OrElse DB Is Nothing Then
                Exit Function
            End If

            'Step 1 - Get the data from the DataTable and put it into an array
            Dim LayoutArray() As Byte = Nothing
            Dim oGridLayoutTable As DataTable = DB.SelectLayoutDataTable(UniqueContainerName, Me.Name)

            If oGridLayoutTable Is Nothing OrElse oGridLayoutTable.Rows.Count = 0 Then
                Exit Function
            End If


            'Get layout
            LayoutArray = oGridLayoutTable.Rows(0)("Layout")

            If LayoutArray.Length = 0 Then
                Exit Function
            End If

            'Step 2 - Create a Memory Stream from the array
            Dim MS As New IO.MemoryStream(LayoutArray)

            'Step 3 - Load the layout from the MemoryStream into the grid
            MS.Seek(0, IO.SeekOrigin.Begin)
            Me.DisplayLayout.Load(MS, Infragistics.Win.UltraWinGrid.PropertyCategories.All)

            _SavedLayout = Me.DisplayLayout.Clone

            Return HasSavedLayout

        Catch ex As Exception
            _ExManager.ShowForm(ex, "Loading layout for: " & Me.UniqueContainerName & Me.Name)
        End Try
    End Function

    ''' <summary>
    ''' Deletes the layout.
    ''' </summary>
    ''' <returns></returns>
    ''' <developer>Pete Sral</developer>
    Public Overridable Function DeleteLayout() As Boolean
        Try

            If MessageBox.Show("Delete saved layout and revert to the default layout?", "Confirm Delete", MessageBoxButtons.YesNo, MessageBoxIcon.Question) = DialogResult.No Then
                Return False
            End If

            If DB.DeleteLayout(Me.UniqueContainerName, Me.Name) Then

                If SavedLayout IsNot Nothing Then
                    _SavedLayout.Dispose()
                    _SavedLayout = Nothing
                End If

                ResetLayout()
                SmartTagMenuState()
                Return True
            Else
                Return False
            End If

        Catch ex As Exception
            _ExManager.ShowForm(ex, "Deleting layout for: " & Me.UniqueContainerName & Me.Name)
        End Try
    End Function

    ''' <summary>
    ''' Handles the layout.
    ''' </summary>
    ''' <developer>Pete Sral</developer>
    Protected Friend Overridable Sub HandleLayout()
        'set the current layout after the grid loads
        If Not HasSavedLayout OrElse ResetLayoutAfterLoad Then
            If LoadLayout() Then
                InitializeCurrentLayout(Me, CurrentLayout)
            End If
        Else
            ResetLayout(SavedLayout)
            InitializeCurrentLayout(Me, SavedLayout)
        End If
    End Sub


#End Region


#Region "Exporting"
    Private _ExcelExportHeaderCount As Integer
    Private _ExcelExportCurrentWorksheet As Integer
    Private _ExcelExportProgress As Integer


    Private WithEvents _ExcelExporter As ExcelExport.UltraGridExcelExporter
    '''' <summary>
    '''' UltraGrid Excel Exporter
    '''' </summary>
    '''' <value></value>
    '''' <returns></returns>
    '''' <remarks></remarks>
    Public Property ExcelExporter() As ExcelExport.UltraGridExcelExporter
        Get
            Return _ExcelExporter
        End Get
        Set(ByVal value As ExcelExport.UltraGridExcelExporter)
            _ExcelExporter = value
        End Set
    End Property


    ''' <summary>
    ''' Export to selected Excel file.
    ''' </summary>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Overridable Function ExportExcel() As String
        Try
            Dim saveFile As New SaveFileDialog()
            'Set the file name filter to default to .xls to list the type of file to save as, and present
            '.xls as the first file type in the drop down list.
            saveFile.Filter = "Excel 2003 (*.xls)|*.xls|Excel 2007-10 (*.xlsx)|*.xlsx|All files (*.*)|*.*"
            saveFile.FilterIndex = 2
            saveFile.DefaultExt = ".xlsx"

            'Open a Windows save file dialog box and ask the user to save the file where ever they want.
            'Pass in the grid and the FileName property to the exporter.

            Using saveFile
                If saveFile.ShowDialog = DialogResult.OK Then

                    If ExportExcel(saveFile.FileName) Then
                        Return saveFile.FileName
                    End If
                End If
            End Using

            Return String.Empty
        Catch ex As Exception
            _ExManager.ShowForm(ex, "Exporting to Excel")
            Return String.Empty
        End Try
    End Function

    ''' <summary>
    ''' Export to Excel file.
    ''' </summary>
    ''' <param name="filename"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Overridable Function ExportExcel(ByVal filename As String) As Boolean
        'Dim oExcel As Infragistics.Win.UltraWinGrid.ExcelExport.UltraGridExcelExporter = Nothing
        Dim oArg As New Bradford.Common.CancelEventArg
        Dim oBook As Workbook
        Dim iterations As Integer
        Dim iSheet As Integer

        If filename.ToLower.EndsWith(".xlsx") Then
            oBook = New Workbook(WorkbookFormat.Excel2007)
            _ExcelMaxRows = Workbook.MaxExcel2007RowCount
            iterations = Me.Rows.Count / _ExcelMaxRows
        Else
            oBook = New Workbook(WorkbookFormat.Excel97To2003)
            _ExcelMaxRows = Workbook.MaxExcelRowCount
            iterations = Me.Rows.Count / _ExcelMaxRows
        End If

        RaiseEvent BeforeExport(filename, Me, oArg)

        If oArg.Cancel Then
            Return False
        End If

        If ProgressBar IsNot Nothing Then
            ProgressBar.Maximum = Me.Rows.Count
        End If

        Try
            Cursor = Cursors.WaitCursor
            If Not String.IsNullOrEmpty(filename) Then

                If Me.Rows.Count > iterations Then
                    _ExcelExportHeaderCount = 0

                    For iSheet = 0 To iterations
                        _ExcelExportCurrentWorksheet = iSheet
                        _ExcelExportProgress = 0

                        Dim oSheet As Worksheet = oBook.Worksheets.Add("Sheet " & (iSheet + 1).ToString & " of " & (iterations + 1).ToString)

                        'oExcel = New Infragistics.Win.UltraWinGrid.ExcelExport.UltraGridExcelExporter
                        If _ExcelExporter.Export(Me, oSheet) Is Nothing Then
                            RaiseEvent ExportAborted(Me, New EventArgs)
                        End If
                    Next

                    oBook.Save(filename)
                Else
                    If _ExcelExporter.Export(Me, filename) Is Nothing Then
                        RaiseEvent ExportAborted(Me, New EventArgs)
                    End If
                End If

                RaiseEvent AfterExport(filename, Me, oArg)

                If Not oArg.Cancel Then
                    Bradford.Common.ShellExecuteAssoc(filename)
                End If

                Return FileIO.FileSystem.FileExists(filename)
            Else
                Return False
            End If
        Catch ex As Exception
            _ExManager.ShowForm(ex, "Excel Export to file: " & filename)
        Finally
            If ProgressBar IsNot Nothing Then
                ProgressBar.Value = 0
            End If
            _ExcelExportProgress = 0
            Cursor = Cursors.Default
        End Try
    End Function

    ''' <summary>
    ''' Create a unique container name.
    ''' </summary>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Overridable Function GenerateUniqueContainerName() As String
        Dim sName As String = String.Empty
        _UniqueContainerName = String.Empty

        Try
            If Not DesignMode Then

                sName = Me.FindForm.Name & "."

                sName &= GetControlPath(Me)

                'limit 
                If sName.Length > 255 Then
                    sName = sName.Substring(255)
                End If

            End If

        Catch ex As Exception
            ' _ExManager.ShowForm(ex, "Generating Unique Container Name")
        End Try

        Return sName

    End Function
    Private Sub _ExcelExporter_CellExporting(ByVal sender As Object, ByVal e As Infragistics.Win.UltraWinGrid.ExcelExport.CellExportingEventArgs) Handles _ExcelExporter.CellExporting
        If GetValue(e.Value).Length > Int16.MaxValue Then
            e.Value = GetValue(e.Value).Substring(0, Int16.MaxValue)
        End If
    End Sub


    Private Sub _ExcelExporter_HeaderRowExporting(ByVal sender As Object, ByVal e As Infragistics.Win.UltraWinGrid.ExcelExport.HeaderRowExportingEventArgs) Handles _ExcelExporter.HeaderRowExporting
        ' increment the header count rows
        _ExcelExportHeaderCount += 1
    End Sub


    Private Sub _ExcelExporter_InitializeRow(ByVal sender As Object, ByVal e As Infragistics.Win.UltraWinGrid.ExcelExport.ExcelExportInitializeRowEventArgs) Handles _ExcelExporter.InitializeRow
        Try


            Dim iTotalRows As Integer = Me.Rows.Count
            Dim iTotalPass As Integer = iTotalRows \ _ExcelMaxRows
            Dim sAction As String = "Exporting"

            If iTotalPass = 0 Then
                iTotalPass = 1
            End If

            'export the rows that should be included to current sheet, based on the header , and calculated Worksheets 
            If Not (_ExcelExportCurrentWorksheet * _ExcelMaxRows < e.Row.Index + _ExcelExportHeaderCount AndAlso e.Row.Index <= (_ExcelExportCurrentWorksheet + _ExcelExportHeaderCount) * _ExcelMaxRows + _ExcelMaxRows) Then
                e.SkipRow = True
                sAction = "Skipping"
            Else
                'determine if there is more than 1 pass.
                If (Me.Rows.Count / _ExcelMaxRows) - iTotalPass > 0 Then
                    iTotalPass += 1

                    'calculate the remainder for the last pass.
                    If iTotalPass = _ExcelExportHeaderCount Then
                        iTotalRows = (Me.Rows.Count - (_ExcelMaxRows * (_ExcelExportHeaderCount - 1))) + 1
                    End If
                End If

                _ExcelExportProgress += 1

                If ProgressBar IsNot Nothing Then
                    If _ExcelExportProgress > ProgressBar.Maximum Then
                        ProgressBar.Maximum = _ExcelExportProgress + 1
                    End If
                    ProgressBar.Value = _ExcelExportProgress
                End If

                If GlobalObject IsNot Nothing Then
                    GlobalObject.TopMessage.Update(sAction & " Row (Pass " & _ExcelExportHeaderCount.ToString & " of " & FormatNumber(iTotalPass, 0, TriState.False) & ") - " & _ExcelExportProgress.ToString & " of " & iTotalRows, "ExportExcel")
                End If

                Application.DoEvents()
            End If

        Catch ex As Exception
            HandleEx(ex, "Excel Exporter - Initialize Row.")
        End Try
    End Sub
#End Region

#Region "Row Selecting"
    ''' <summary>
    ''' First row selected.
    ''' </summary>
    Public ReadOnly Property FirstSelectedRow() As UltraGridRow
        Get
            If SelectedRows Is Nothing OrElse SelectedRows.Count = 0 Then
                Return ActiveRow
            Else
                Return SelectedRows(0)
            End If
        End Get
    End Property
    Private _SelectedRows As List(Of UltraGridRow)
    ''' <summary>
    ''' Custom selected rows list.
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public ReadOnly Property SelectedRows() As List(Of UltraGridRow)
        Get
            Return _SelectedRows
        End Get
    End Property

    ''' <summary>
    ''' Count of rows in custom selected list including the ActiveRow (if not in selected list).
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Overridable ReadOnly Property SelectedRowsCount() As Integer
        Get

            Return Me.Selected.Rows.Count

            'Dim iCount As Integer = SelectedRows.Count

            'If Me.ActiveRow IsNot Nothing Then
            '    If Not SelectedRows.Contains(Me.ActiveRow) Then
            '        iCount += 1
            '    End If
            'End If

            'For Each row As UltraGridRow In Me.Selected.Rows
            '    If Not SelectedRows.Contains(row) Then
            '        iCount += 1
            '    End If
            'Next
            'Return iCount
        End Get

    End Property

    Private Function GetRowSelectorElement(ByVal e As System.Windows.Forms.MouseEventArgs) As RowSelectorUIElement
        Dim oElement As UIElement
        Dim oRowSelectorElement As RowSelectorUIElement

        oElement = Me.DisplayLayout.UIElement.ElementFromPoint(New Point(e.X, e.Y))
        If Not oElement Is Nothing Then

            oRowSelectorElement = oElement.GetContext(GetType(RowSelectorUIElement))

            If oRowSelectorElement Is Nothing Then
                oRowSelectorElement = oElement.GetAncestor(GetType(RowSelectorUIElement))
            End If

            If oRowSelectorElement Is Nothing Then
                Return Nothing
            Else
                Return oRowSelectorElement
            End If
        End If

        Return Nothing

    End Function

    ''' <summary>
    ''' Select the row when clicking on row selector cell.
    ''' </summary>
    ''' <param name="e"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Function GetRowOnSelectorClick(ByVal e As System.Windows.Forms.MouseEventArgs) As UltraGridRow
        Try
            Dim oPoint As New Point(e.X, e.Y)

            Dim oElement As UIElement = GetRowSelectorElement(e)

            If oElement IsNot Nothing Then
                Return TryCast(oElement.GetContext(GetType(UltraGridRow)), UltraGridRow)
            Else
                Return Nothing
            End If

        Catch ex As Exception
            Throw
        End Try
    End Function

    ''' <summary>
    ''' Add row to SelectedRows list with Before and After events.
    ''' </summary>
    ''' <param name="row"></param>
    ''' <remarks></remarks>
    Protected Friend Overridable Sub AddSelectedRow(ByVal row As UltraGridRow)
        Try

            'skip if not in multi row select mode.
            If Me.CurrentLayout.Override.SelectTypeRow = SelectType.Single Or _
               Me.CurrentLayout.Override.SelectTypeRow = SelectType.SingleAutoDrag Or _
               Me.CurrentLayout.Override.SelectTypeRow = SelectType.None Then
                Exit Sub
            End If

            OnBeforeSelectRow(row)

            SelectedRows.Add(row)

            If Not Me.Selected.Rows.Contains(row) Then
                Me.Selected.Rows.Add(row)
            End If

            OnAfterSelectedRow(row)
        Catch ex As Exception
            HandleEx(ex, "Adding to selected rows list.")
        End Try
    End Sub

    ''' <summary>
    ''' Remove row from SelectedRows list with Before and After events.
    ''' </summary>
    ''' <param name="row"></param>
    ''' <remarks></remarks>
    Protected Friend Overridable Sub RemoveSelectedRow(ByVal row As UltraGridRow)
        Try
            OnBeforeRemoveRow(row)

            SelectedRows.Remove(row)

            If Me.Selected.Rows.Contains(row) Then
                Me.Selected.Rows.Clear()
            End If

            OnAfterRemoveRow(row)
        Catch ex As Exception
            HandleEx(ex, "Removing from selected rows list.")
        End Try
    End Sub

    ''' <summary>
    ''' Removed active row from selected rows and repaint row.
    ''' </summary>
    ''' <remarks></remarks>
    Public Overridable Sub UnselectRow()
        If Not Me.ActiveRow Is Nothing Then
            UnselectRow(Me.ActiveRow)
        End If
    End Sub
    ''' <summary>
    ''' Remove row from selected rows and repaint row.
    ''' </summary>
    ''' <param name="row"></param>
    ''' <remarks></remarks>
    Public Overridable Sub UnselectRow(ByVal row As UltraGridRow)
        If Me.SelectedRows.Contains(row) Then
            RemoveSelectedRow(row)
            ReInitializeRow(row)
        End If
    End Sub
    ''' <summary>
    ''' Remove row matching data row view from selected rows and reprint.
    ''' </summary>
    ''' <param name="row"></param>
    ''' <remarks></remarks>
    Public Overridable Sub UnselectRow(ByVal row As DataRowView, ByVal columnName As String)
        If Me.SelectRow(columnName, row(columnName)) Then
            RemoveSelectedRow(Me.ActiveRow)
            ReInitializeRow()
        End If
    End Sub

    ''' <summary>
    ''' FireInitilizeRow for row matching data row view to repaint.
    ''' </summary>
    ''' <param name="row"></param>
    ''' <remarks></remarks>
    Public Overridable Sub ReInitializeRow(ByVal row As DataRowView, ByVal columnName As String)
        If Me.SelectRow(columnName, row(columnName)) Then
            ReInitializeRow()
        End If
    End Sub
    ''' <summary>
    ''' FireInitilizeRow for ActiveRow to repaint.
    ''' </summary>
    ''' <remarks></remarks>
    Public Overridable Sub ReInitializeRow()
        If Not Me.ActiveRow Is Nothing Then
            ReInitializeRow(Me.ActiveRow)
        End If
    End Sub
    ''' <summary>
    ''' FireInitilizeRow for row  to repaint.
    ''' </summary>
    ''' <param name="row"></param>
    ''' <remarks></remarks>
    Public Overridable Sub ReInitializeRow(ByVal row As UltraGridRow)
        If Not row Is Nothing Then
            row.Refresh(UltraWinGrid.RefreshRow.FireInitializeRow)
        End If
    End Sub

    ''' <summary>
    ''' Initialize each row in grid.
    ''' </summary>
    ''' <developer>Pete Sral</developer>
    Public Overridable Sub ReInitializeRows()
        Me.SuspendLayout()
        For Each row As UltraGridRow In Me.Rows
            ReInitializeRow(row)
        Next
        Me.ResumeLayout()
    End Sub

    ''' <summary>
    ''' Select next document row after ActiveRow
    ''' </summary>
    ''' <remarks></remarks>
    Public Overridable Sub SelectNextRow()

        Try

            If ActiveRow Is Nothing Then
                Exit Sub
            End If

            Dim iParentIndex As Integer = -1
            Dim iRowIndex As Integer = -1


            If ActiveRow.ParentRow IsNot Nothing Then
                iParentIndex = ActiveRow.ParentRow.Index
            End If
            iRowIndex = ActiveRow.Index

            SelectNextRow(iParentIndex, iRowIndex)
        Catch ex As Exception
            HandleEx(ex, "Selecting next row from active row.")
        End Try

    End Sub
    ''' <summary>
    ''' Select next document row.
    ''' </summary>
    ''' <param name="parentIndex">Original Parent Index (-1 if not in a branch)</param>
    ''' <param name="rowIndex">Original Row Index</param>
    ''' <remarks></remarks>
    Public Overridable Sub SelectNextRow(ByVal parentIndex As Integer, ByVal rowIndex As Integer)
        Try

            If parentIndex = -1 And rowIndex = -1 Then
                'no active row
                Exit Sub
            End If

            If parentIndex = -1 Then
                'are we at the end
                If Me.Rows.Count - 1 < rowIndex + 1 And Not Me.Rows(rowIndex).HasChild Then
                    'on last row and is not a branch
                    Exit Sub
                ElseIf (Me.Rows.Count - 1 < rowIndex + 1 And Me.Rows(rowIndex).HasChild) OrElse Me.Rows(rowIndex).HasChild Then
                    'on last row and it is a branch, select the branch
                    SelectNextRow(ActiveRow.Index, -1)
                Else
                    Me.Rows(rowIndex + 1).Activate()
                    Me.Rows(rowIndex + 1).Selected = True
                End If
            Else 'inside band
                If Me.Rows(parentIndex).ChildBands(0).Rows.Count - 1 < rowIndex + 1 Then
                    'at the end of the band
                    If Me.Rows.Count - 1 < parentIndex + 1 Then
                        'at the end of grid
                        Exit Sub
                    Else
                        Me.Rows(parentIndex + 1).Activate()
                        Me.Rows(parentIndex + 1).Selected = True
                    End If
                Else
                    Dim iNextRowIndex As Integer = rowIndex + 1

                    ''still inside band
                    ''handle reposition of row but the new index will always be on top
                    Me.Rows(parentIndex).ChildBands(0).Rows(iNextRowIndex).Activate()
                    Me.Rows(parentIndex).ChildBands(0).Rows(iNextRowIndex).Selected = True
                End If
            End If


            'on a branch row, need to go inside.
            If Not ActiveRow Is Nothing Then
                If ActiveRow.HasChild Then
                    SelectNextRow(ActiveRow.Index, -1)
                End If
            End If


        Catch ex As Exception
            HandleEx(ex, "Selecting next row after Parent Index: " & parentIndex & ", Row Index: " & rowIndex)
        End Try
    End Sub


    ''' <summary>
    ''' Display or Remove check box/row coloring for selected/unselected items.
    ''' </summary>
    ''' <param name="row"></param>
    ''' <remarks></remarks>
    Protected Friend Overridable Sub ColorSelectedRow(ByVal row As UltraGridRow)

        If SelectedRows.Contains(row) Then
            If GlobalObject IsNot Nothing Then
                row.Appearance.BackColor = GlobalObject.Settings.GetSetting("GridSelectedRowColor", Color.LightYellow)
            Else
                row.Appearance.BackColor = Color.LightYellow
            End If

            Dim sImageKey As String = "Check"
            Dim checkOne As Boolean
            If row.Equals(SelectedRows.Item(0)) Then
                sImageKey = "Check1"
                checkOne = True
            End If

            If GlobalObject IsNot Nothing Then
                row.RowSelectorAppearance.Image = GlobalObject.ImageResources.GetImage(sImageKey)
            Else
                If checkOne Then
                    row.RowSelectorAppearance.Image = My.Resources.check1
                Else
                    row.RowSelectorAppearance.Image = My.Resources.check
                End If
            End If
        Else
            row.Appearance.ResetBackColor()
            row.RowSelectorAppearance.ResetImage()

            If SelectedRows.Count <> 0 Then
                Dim oRow As UltraGridRow = SelectedRows(0)

                If oRow IsNot Nothing Then
                    If GlobalObject IsNot Nothing Then
                        oRow.RowSelectorAppearance.Image = GlobalObject.ImageResources.GetImage("Check1")
                    Else
                        row.RowSelectorAppearance.Image = My.Resources.check1
                    End If
                End If
            End If
        End If

    End Sub
    ''' <summary>
    ''' Handle the logic to select a row.
    ''' </summary>
    ''' <param name="row"></param>
    ''' <remarks></remarks>
    Protected Friend Overridable Sub HandleRowSelect(ByVal row As UltraGridRow)
        If row Is Nothing Then
            Exit Sub
        End If


        If SelectedRows.Contains(row) Then
            RemoveSelectedRow(row)

        Else
            AddSelectedRow(row)

        End If

        ColorSelectedRow(row)


    End Sub

#End Region

#Region "Inline Messages"
    ''' <summary>
    ''' Inline message counts
    ''' </summary>
    ''' <returns></returns>
    ''' <developer>Pete Sral</developer>
    Private Function MessageCount() As Integer
        Dim iCount As Integer
        For Each ctl As Control In Me.Controls
            If TypeOf (ctl) Is BSCUltraDataGridMessage Then
                iCount += 1
            End If
        Next

        Return iCount - 1

    End Function
    ''' <summary>
    ''' Hides the inline messages.
    ''' </summary>
    ''' <developer>Pete Sral</developer>
    Public Sub HideMessages()
        Dim oMsgCtl As BSCUltraDataGridMessage
        For Each ctl As Control In Me.Controls
            If TypeOf (ctl) Is BSCUltraDataGridMessage Then
                Me.Controls.Remove(ctl)
                oMsgCtl = DirectCast(ctl, BSCUltraDataGridMessage)
                oMsgCtl.Close()
                oMsgCtl.Dispose()
                oMsgCtl = Nothing
            End If
        Next
    End Sub
    ''' <summary>
    ''' Shows the inline message.
    ''' </summary>
    ''' <param name="message">The message.</param>
    ''' <param name="image">The image.</param>
    ''' <param name="overrideAutoClose">if set to <c>true</c> override auto close.</param>
    ''' <developer>Pete Sral</developer>
    Public Sub ShowMessage(ByVal message As String, Optional ByVal image As Image = Nothing, Optional overrideAutoClose As Boolean = False)
        Dim oMsgCtl As New BSCUltraDataGridMessage
        Dim iMessageCount As Integer

        With oMsgCtl

            .BorderStyle = BorderStyle.Fixed3D
            If Not overrideAutoClose Then
                .AutoClose = GlobalObject.Settings.GetSetting("GridMsgAutoClose", True)
                .AutoCloseTime = GlobalObject.Settings.GetSetting("GridMsgAutoCloseTime", 5)
            End If

            Me.Controls.Add(oMsgCtl)
            'position
            .Left = 1  'Me.Left + Me.Width - _SmartTag.Width
            iMessageCount = MessageCount()
            If iMessageCount <= 0 Then
                .Top = 1
            Else
                .Top = ((oMsgCtl.Height) * iMessageCount) + 1
            End If

            .Width = Me.Width - 1

            .Message = message
            If image IsNot Nothing Then
                .Image = image
            End If

            .Display()
        End With
    End Sub

#End Region

#Region "Dispose"
    Private _disposedValue As Boolean

    ''' <summary>
    ''' Gets a value indicating whether [disposed value].
    ''' </summary>
    ''' <value><c>true</c> if [disposed value]; otherwise, <c>false</c>.</value>
    ''' <developer>Pete Sral</developer>
    Protected ReadOnly Property DisposedValue() As Boolean
        Get
            Return _disposedValue
        End Get
    End Property

    ''' <summary>
    ''' Disposes the objects.
    ''' </summary>
    ''' <param name="disposing">if set to <c>true</c> [disposing].</param>
    ''' <developer>Pete Sral</developer>
    Protected Overridable Sub DisposeObjects(ByVal disposing As Boolean)
        If DefaultLayout IsNot Nothing Then
            DefaultLayout.Dispose()
            _DefaultLayout = Nothing
        End If

        If SavedLayout IsNot Nothing Then
            SavedLayout.Dispose()
            _SavedLayout = Nothing
        End If

        If _DB IsNot Nothing Then
            _DB.Dispose()
        End If

        _ExManager = Nothing

        _HiddenColumns = Nothing

        If SmartTagMenuLeft IsNot Nothing AndAlso Not SmartTagMenuLeft.IsDisposed Then
            SmartTagMenuLeft.Dispose()
            _SmartTagMenuLeft = Nothing
        End If

        If ExcelExporter IsNot Nothing Then
            ExcelExporter.Dispose()
            _ExcelExporter = Nothing
        End If

    End Sub
    Protected Overrides Sub Dispose(ByVal disposing As Boolean)
        SyncLock Me
            Try
                'Debug.Print(">>>>> " & Now.ToLongTimeString & " Disposing: " & _UnqiueContainerName)
                If disposing AndAlso components IsNot Nothing Then
                    components.Dispose()
                End If

                If Not _disposedValue Then
                    If disposing Then
                        ' free other state (managed objects).
                        DisposeObjects(disposing)
                    End If

                    '  free your own state (unmanaged objects).
                    '  set large fields to null.
                End If
                _disposedValue = True
            Finally
                MyBase.Dispose(disposing)
            End Try
        End SyncLock
    End Sub
#End Region
End Class
